"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.onTreeChanged = exports.onNodeChanged = void 0;
/**
 * Subscribe to changes to the node for the given {@link AnchorNode}.
 * @remarks This fulfills the contract of {@link TreeChangeEvents.nodeChanged}.
 * @privateRemarks The logic in this function ensures that the listener is only fired once per change per node.
 */
function onNodeChanged(anchorNode, listener) {
    // Debounce "childrenChanged" (which fires separately for each field that changed in the node)
    // by waiting for "subtreeChanged" (which only fires once regardless of how many fields changed).
    let unsubscribeFromTreeChanged;
    // Every time that "childrenChanged" fires...
    return anchorNode.on("childrenChanged", () => {
        // ...subscribe to "subtreeChanged", but only if we haven't subscribed already already since the last time it fired
        if (unsubscribeFromTreeChanged === undefined) {
            unsubscribeFromTreeChanged = anchorNode.on("subtreeChanged", () => {
                listener();
                unsubscribeFromTreeChanged?.();
                unsubscribeFromTreeChanged = undefined;
            });
        }
    });
}
exports.onNodeChanged = onNodeChanged;
/**
 * Subscribe to changes to the tree rooted at the given {@link AnchorNode}.
 * @remarks This fulfills the contract of {@link TreeChangeEvents.treeChanged}.
 */
function onTreeChanged(anchorNode, listener) {
    return anchorNode.on("subtreeChanged", listener);
}
exports.onTreeChanged = onTreeChanged;
//# sourceMappingURL=treeEvents.js.map