"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.configuredSharedTree = exports.SharedTree = exports.TreeFactory = void 0;
const internal_1 = require("@fluidframework/shared-object-base/internal");
const packageVersion_js_1 = require("./packageVersion.js");
const index_js_1 = require("./shared-tree/index.js");
/**
 * A channel factory that creates an {@link ITree}.
 */
class TreeFactory {
    options;
    static Type = "https://graph.microsoft.com/types/tree";
    static attributes = {
        type: this.Type,
        snapshotFormatVersion: "0.0.0",
        packageVersion: packageVersion_js_1.pkgVersion,
    };
    type = TreeFactory.Type;
    attributes = TreeFactory.attributes;
    constructor(options) {
        this.options = options;
    }
    async load(runtime, id, services, channelAttributes) {
        const tree = new index_js_1.SharedTree(id, runtime, channelAttributes, this.options);
        await tree.load(services);
        return tree;
    }
    create(runtime, id) {
        const tree = new index_js_1.SharedTree(id, runtime, this.attributes, this.options);
        tree.initializeLocal();
        return tree;
    }
}
exports.TreeFactory = TreeFactory;
/**
 * SharedTree is a hierarchical data structure for collaboratively editing strongly typed JSON-like trees
 * of objects, arrays, and other data types.
 * @alpha
 */
exports.SharedTree = configuredSharedTree({});
/**
 * {@link SharedTree} but allowing a non-default configuration.
 * @remarks
 * This is useful for debugging and testing to opt into extra validation or see if opting out of some optimizations fixes an issue.
 * @example
 * ```typescript
 * import {
 * 	ForestType,
 * 	TreeCompressionStrategy,
 * 	configuredSharedTree,
 * 	typeboxValidator,
 * 	// eslint-disable-next-line import/no-internal-modules
 * } from "@fluidframework/tree/internal";
 * const SharedTree = configuredSharedTree({
 * 	forest: ForestType.Reference,
 * 	jsonValidator: typeboxValidator,
 * 	treeEncodeType: TreeCompressionStrategy.Uncompressed,
 * });
 * ```
 * @privateRemarks
 * TODO:
 * Expose Ajv validator for better error message quality somehow.
 * Maybe as part of a test utils or dev-tool package?
 * @internal
 */
function configuredSharedTree(options) {
    class ConfiguredFactory extends TreeFactory {
        constructor() {
            super(options);
        }
    }
    return (0, internal_1.createSharedObjectKind)(ConfiguredFactory);
}
exports.configuredSharedTree = configuredSharedTree;
//# sourceMappingURL=treeFactory.js.map